!###############################################################################
!#                                                                             #
!# aed_phytoplankton.F90                                                       #
!#                                                                             #
!#  Developed by :                                                             #
!#      AquaticEcoDynamics (AED) Group                                         #
!#      The University of Western Australia                                    #
!#                                                                             #
!#      http://aquatic.science.uwa.edu.au/                                     #
!#                                                                             #
!#  Copyright 2013 - 2022 -  The University of Western Australia               #
!#                                                                             #
!#   AED is free software: you can redistribute it and/or modify               #
!#   it under the terms of the GNU General Public License as published by      #
!#   the Free Software Foundation, either version 3 of the License, or         #
!#   (at your option) any later version.                                       #
!#                                                                             #
!#   AED is distributed in the hope that it will be useful,                    #
!#   but WITHOUT ANY WARRANTY; without even the implied warranty of            #
!#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             #
!#   GNU General Public License for more details.                              #
!#                                                                             #
!#   You should have received a copy of the GNU General Public License         #
!#   along with this program.  If not, see <http://www.gnu.org/licenses/>.     #
!#                                                                             #
!#   -----------------------------------------------------------------------   #
!#                                                                             #
!# Created August 2011                                                         #
!#                                                                             #
!#  Track changes on GitHub @ https://github.com/AquaticEcoDynamics/libaed-water
!#                                                                             #
!###############################################################################
!                                                                              !
!         .----------------.  .----------------.  .----------------.           !
!         | .--------------. || .--------------. || .--------------. |         !
!         | |   ______     | || |  ____  ____  | || |  ____  ____  | |         !
!         | |  |_   __ \   | || | |_   ||   _| | || | |_  _||_  _| | |         !
!         | |    | |__) |  | || |   | |__| |   | || |   \ \  / /   | |         !
!         | |    |  ___/   | || |   |  __  |   | || |    \ \/ /    | |         !
!         | |   _| |_      | || |  _| |  | |_  | || |    _|  |_    | |         !
!         | |  |_____|     | || | |____||____| | || |   |______|   | |         !
!         | |              | || |              | || |              | |         !
!         | '--------------' || '--------------' || '--------------' |         !
!         '----------------'  '----------------'  '----------------'           !
!                                                                              !
!###############################################################################

#include "aed.h"

MODULE aed_phytoplankton
!-------------------------------------------------------------------------------
!  aed_phytoplankton --- phytoplankton biogeochemical model
!-------------------------------------------------------------------------------
   USE aed_core
   USE aed_util
   USE aed_bio_utils

   IMPLICIT NONE

   PRIVATE   ! By default make everything private

   PUBLIC aed_phytoplankton_data_t

   TYPE,extends(aed_model_data_t) :: aed_phytoplankton_data_t
      !# Variable identifiers
      INTEGER,ALLOCATABLE :: id_p(:)
      INTEGER,ALLOCATABLE :: id_in(:)
      INTEGER,ALLOCATABLE :: id_ip(:)
      INTEGER,ALLOCATABLE :: id_rho(:)
      INTEGER,ALLOCATABLE :: id_NtoP(:)
      INTEGER,ALLOCATABLE :: id_vvel(:)
      INTEGER,ALLOCATABLE :: id_fT(:), id_fI(:), id_fNit(:), &
                             id_fPho(:), id_fSil(:), id_fSal(:)

      INTEGER,ALLOCATABLE :: id_PhyGPPc(:),id_PhyRSPc(:),id_PhyEXCc(:),id_PhyMORc(:),id_PhySEDc(:)
      INTEGER,ALLOCATABLE :: id_PhyGPPn(:),id_PhyRSPn(:),id_PhyEXCn(:),id_PhyMORn(:),id_PhySEDn(:)
      INTEGER,ALLOCATABLE :: id_PhyGPPp(:),id_PhyRSPp(:),id_PhyEXCp(:),id_PhyMORp(:),id_PhySEDp(:)

      INTEGER :: id_Pexctarget,id_Pmorttarget,id_Pupttarget(1:2)
      INTEGER :: id_Nexctarget,id_Nmorttarget,id_Nupttarget(1:4)
      INTEGER :: id_Cexctarget,id_Cmorttarget,id_Cupttarget
      INTEGER :: id_Siexctarget,id_Simorttarget,id_Siupttarget
      INTEGER :: id_DOupttarget, id_l_resus, id_Psed_phy
      INTEGER :: id_par, id_I_0, id_extc, id_sedzone
      INTEGER :: id_tem, id_sal, id_dz, id_dens
      INTEGER :: id_GPP, id_NCP, id_PPR, id_NPR, id_dPAR
      INTEGER :: id_TPHY,id_chl, id_TIN, id_TIP
      INTEGER :: id_MPB, id_d_MPB, id_d_BPP, id_d_BCP, id_d_mpbv, id_d_res
      INTEGER :: id_NUP1,id_NUP2,  id_NUP3,  id_NUP4,  id_PUP,    id_CUP

      !# Model parameters
      INTEGER  :: num_phytos
      TYPE(phyto_data_t),DIMENSION(:),ALLOCATABLE :: phytos
      ! LOGICAL  :: do_exc,do_mort,do_upt, do_N2uptake
      LOGICAL  :: do_Puptake, do_Nuptake, do_Cuptake
      LOGICAL  :: do_Siuptake, do_DOuptake, do_N2uptake
      LOGICAL  :: do_Pmort, do_Nmort, do_Cmort, do_Simort
      LOGICAL  :: do_Pexc, do_Nexc, do_Cexc, do_Siexc
      INTEGER  :: do_mpb, n_zones
      AED_REAL :: R_mpbg, R_mpbr, I_Kmpb, mpb_max, theta_mpb_growth, theta_mpb_resp
      INTEGER  :: nnup, npup
      AED_REAL :: dic_per_n
      AED_REAL :: min_rho,max_rho
      AED_REAL,ALLOCATABLE :: resuspension(:), active_zones(:)

     CONTAINS
         PROCEDURE :: define            => aed_define_phytoplankton
         PROCEDURE :: calculate         => aed_calculate_phytoplankton
         PROCEDURE :: calculate_benthic => aed_calculate_benthic_phytoplankton
         PROCEDURE :: mobility          => aed_mobility_phytoplankton
         PROCEDURE :: light_extinction  => aed_light_extinction_phytoplankton
!        PROCEDURE :: delete            => aed_delete_phytoplankton

   END TYPE

! MODULE GLOBALS
   AED_REAL :: dtlim = 0.9 * 3600
   LOGICAL  :: extra_diag = .false.
   INTEGER  :: diag_level = 10                ! 0 = no diagnostic outputs
                                              ! 1 = basic diagnostic outputs
                                              ! 2 = flux rates, and supporitng
                                              ! 3 = other metrics
                                              !10 = all debug & checking outputs

!===============================================================================
CONTAINS


!###############################################################################
INTEGER FUNCTION load_csv(dbase,pd)
!-------------------------------------------------------------------------------
   USE aed_csv_reader
!-------------------------------------------------------------------------------
!ARGUMENTS
   CHARACTER(len=*),INTENT(in) :: dbase
   TYPE(phyto_param_t) :: pd(MAX_PHYTO_TYPES)
!
!LOCALS
   INTEGER :: unit, nccols, ccol, dcol
   CHARACTER(len=32),POINTER,DIMENSION(:) :: csvnames
   CHARACTER(len=32) :: name
   TYPE(AED_SYMBOL),DIMENSION(:),ALLOCATABLE :: values
   INTEGER :: idx_col = 0
   LOGICAL :: meh
   INTEGER :: ret = 0
!
!BEGIN
!-------------------------------------------------------------------------------
   unit = aed_csv_read_header(dbase, csvnames, nccols)
   IF (unit <= 0) THEN
      load_csv = -1
      RETURN !# No file found
   ENDIF

   ALLOCATE(values(nccols))

   DO WHILE ( aed_csv_read_row(unit, values) )
      DO ccol=2,nccols
         dcol = ccol - 1
         pd(dcol)%p_name = csvnames(ccol)

         CALL copy_name(values(1), name)
         SELECT CASE (name)
            CASE ('p0')            ; pd(dcol)%p0            = extract_double(values(ccol))
            CASE ('w_p')           ; pd(dcol)%w_p           = extract_double(values(ccol))
            CASE ('Xcc')           ; pd(dcol)%Xcc           = extract_double(values(ccol))
            CASE ('R_growth')      ; pd(dcol)%R_growth      = extract_double(values(ccol))
            CASE ('fT_Method')     ; pd(dcol)%fT_Method     = extract_integer(values(ccol))
            CASE ('theta_growth')  ; pd(dcol)%theta_growth  = extract_double(values(ccol))
            CASE ('T_std')         ; pd(dcol)%T_std         = extract_double(values(ccol))
            CASE ('T_opt')         ; pd(dcol)%T_opt         = extract_double(values(ccol))
            CASE ('T_max')         ; pd(dcol)%T_max         = extract_double(values(ccol))
            CASE ('lightModel')    ; pd(dcol)%lightModel    = extract_integer(values(ccol))
            CASE ('I_K')           ; pd(dcol)%I_K           = extract_double(values(ccol))
            CASE ('I_S')           ; pd(dcol)%I_S           = extract_double(values(ccol))
            CASE ('KePHY')         ; pd(dcol)%KePHY         = extract_double(values(ccol))
            CASE ('f_pr')          ; pd(dcol)%f_pr          = extract_double(values(ccol))
            CASE ('R_resp')        ; pd(dcol)%R_resp        = extract_double(values(ccol))
            CASE ('theta_resp')    ; pd(dcol)%theta_resp    = extract_double(values(ccol))
            CASE ('k_fres')        ; pd(dcol)%k_fres        = extract_double(values(ccol))
            CASE ('k_fdom')        ; pd(dcol)%k_fdom        = extract_double(values(ccol))
            CASE ('salTol')        ; pd(dcol)%salTol        = extract_integer(values(ccol))
            CASE ('S_bep')         ; pd(dcol)%S_bep         = extract_double(values(ccol))
            CASE ('S_maxsp')       ; pd(dcol)%S_maxsp       = extract_double(values(ccol))
            CASE ('S_opt')         ; pd(dcol)%S_opt         = extract_double(values(ccol))
            CASE ('simDINUptake')  ; pd(dcol)%simDINUptake  = extract_integer(values(ccol))
            CASE ('simDONUptake')  ; pd(dcol)%simDONUptake  = extract_integer(values(ccol))
            CASE ('simNFixation')  ; pd(dcol)%simNFixation  = extract_integer(values(ccol))
            CASE ('simINDynamics') ; pd(dcol)%simINDynamics = extract_integer(values(ccol))
            CASE ('N_o')           ; pd(dcol)%N_o           = extract_double(values(ccol))
            CASE ('K_N')           ; pd(dcol)%K_N           = extract_double(values(ccol))
            CASE ('X_ncon')        ; pd(dcol)%X_ncon        = extract_double(values(ccol))
            CASE ('X_nmin')        ; pd(dcol)%X_nmin        = extract_double(values(ccol))
            CASE ('X_nmax')        ; pd(dcol)%X_nmax        = extract_double(values(ccol))
            CASE ('R_nuptake')     ; pd(dcol)%R_nuptake     = extract_double(values(ccol))
            CASE ('k_nfix')        ; pd(dcol)%k_nfix        = extract_double(values(ccol))
            CASE ('R_nfix')        ; pd(dcol)%R_nfix        = extract_double(values(ccol))
            CASE ('simDIPUptake')  ; pd(dcol)%simDIPUptake  = extract_integer(values(ccol))
            CASE ('simIPDynamics') ; pd(dcol)%simIPDynamics = extract_integer(values(ccol))
            CASE ('P_0')           ; pd(dcol)%P_0           = extract_double(values(ccol))
            CASE ('K_P')           ; pd(dcol)%K_P           = extract_double(values(ccol))
            CASE ('X_pcon')        ; pd(dcol)%X_pcon        = extract_double(values(ccol))
            CASE ('X_pmin')        ; pd(dcol)%X_pmin        = extract_double(values(ccol))
            CASE ('X_pmax')        ; pd(dcol)%X_pmax        = extract_double(values(ccol))
            CASE ('R_puptake')     ; pd(dcol)%R_puptake     = extract_double(values(ccol))
            CASE ('simSiUptake')   ; pd(dcol)%simSiUptake   = extract_integer(values(ccol))
            CASE ('Si_0')          ; pd(dcol)%Si_0          = extract_double(values(ccol))
            CASE ('K_Si')          ; pd(dcol)%K_Si          = extract_double(values(ccol))
            CASE ('X_sicon')       ; pd(dcol)%X_sicon       = extract_double(values(ccol))

            CASE DEFAULT ; print *, 'Unknown row "', TRIM(name), '"'
         END SELECT
      ENDDO
   ENDDO

   meh = aed_csv_close(unit)
   !# don't care if close fails

   IF (ASSOCIATED(csvnames)) DEALLOCATE(csvnames)
   IF (ALLOCATED(values))    DEALLOCATE(values)

   load_csv = ret
END FUNCTION load_csv
!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


!###############################################################################
SUBROUTINE aed_phytoplankton_load_params(data, dbase, count, list, settling, resuspension)
!-------------------------------------------------------------------------------
!ARGUMENTS
   CLASS (aed_phytoplankton_data_t),INTENT(inout) :: data
   CHARACTER(len=*),INTENT(in) :: dbase
   INTEGER,INTENT(in)          :: count
   INTEGER,INTENT(in)          :: list(*)
   INTEGER,INTENT(in)          :: settling(*)
   AED_REAL,INTENT(in)         :: resuspension(*)
!
!LOCALS
   INTEGER  :: status
   INTEGER  :: i,tfil
   AED_REAL :: minNut

   TYPE(phyto_param_t) :: pd(MAX_PHYTO_TYPES)
   NAMELIST /phyto_data/ pd     ! %% type phyto_param_t - see aed_bio_utils
!-------------------------------------------------------------------------------
!BEGIN
    SELECT CASE (param_file_type(dbase))
       CASE (CSV_TYPE)
           status = load_csv(dbase, pd)
       CASE (NML_TYPE)
           print*,"nml format parameter file is deprecated. Please update to CSV format"
           tfil = find_free_lun()
           open(tfil,file=dbase, status='OLD',iostat=status)
           IF (status /= 0) STOP 'Cannot open phyto_data namelist file'
           read(tfil,nml=phyto_data,iostat=status)
           close(tfil)
       CASE DEFAULT
           print *,'Unknown file type "',TRIM(dbase),'"'; status=1
    END SELECT
    IF (status /= 0) STOP 'Error reading namelist phyto_data'

    data%num_phytos = count
    ALLOCATE(data%phytos(count))
    ALLOCATE(data%id_p(count)) ; data%id_p(:) = 0
    ALLOCATE(data%id_in(count)) ; data%id_in(:) = 0
    ALLOCATE(data%id_ip(count)) ; data%id_ip(:) = 0
    ALLOCATE(data%id_rho(count)) ; data%id_rho(:) = 0
    ALLOCATE(data%id_NtoP(count)) ; data%id_NtoP(:) = 0
    IF ( diag_level >= 10 ) THEN
       ALLOCATE(data%id_fT(count)) ; data%id_fT(:) = 0
       ALLOCATE(data%id_fI(count)) ; data%id_fI(:) = 0
       ALLOCATE(data%id_fNit(count)) ; data%id_fNit(:) = 0
       ALLOCATE(data%id_fPho(count)) ; data%id_fPho(:) = 0
       ALLOCATE(data%id_fSil(count)) ; data%id_fSil(:) = 0
       ALLOCATE(data%id_fSal(count)) ; data%id_fSal(:) = 0
       ALLOCATE(data%id_vvel(count)) ; data%id_vvel(:) = 0
       ALLOCATE(data%id_PhyGPPc(count)) ; data%id_PhyGPPc(:) = 0
       ALLOCATE(data%id_PhyRSPc(count)) ; data%id_PhyRSPc(:) = 0
       ALLOCATE(data%id_PhyEXCc(count)) ; data%id_PhyEXCc(:) = 0
       ALLOCATE(data%id_PhyMORc(count)) ; data%id_PhyMORc(:) = 0
       ALLOCATE(data%id_PhySEDc(count)) ; data%id_PhySEDc(:) = 0
       ALLOCATE(data%id_PhyGPPn(count)) ; data%id_PhyGPPn(:) = 0
       ALLOCATE(data%id_PhyRSPn(count)) ; data%id_PhyRSPn(:) = 0
       ALLOCATE(data%id_PhyEXCn(count)) ; data%id_PhyEXCn(:) = 0
       ALLOCATE(data%id_PhyMORn(count)) ; data%id_PhyMORn(:) = 0
       ALLOCATE(data%id_PhySEDn(count)) ; data%id_PhySEDn(:) = 0
       ALLOCATE(data%id_PhyGPPp(count)) ; data%id_PhyGPPp(:) = 0
       ALLOCATE(data%id_PhyRSPp(count)) ; data%id_PhyRSPp(:) = 0
       ALLOCATE(data%id_PhyEXCp(count)) ; data%id_PhyEXCp(:) = 0
       ALLOCATE(data%id_PhyMORp(count)) ; data%id_PhyMORp(:) = 0
       ALLOCATE(data%id_PhySEDp(count)) ; data%id_PhySEDp(:) = 0
    ENDIF

    DO i=1,count
       ! Assign parameters from database to chosen simulated groups
       ! Note:  all rates are provided in values per day,
       !        and are converted in here to values per second.
       data%phytos(i)%p_name       = pd(list(i))%p_name
       data%phytos(i)%p0           = pd(list(i))%p0
       data%phytos(i)%w_p          = pd(list(i))%w_p/secs_per_day
       data%phytos(i)%settling     = settling(i)
       data%phytos(i)%resuspension = resuspension(i)
       data%phytos(i)%Xcc          = pd(list(i))%Xcc
       data%phytos(i)%R_growth     = pd(list(i))%R_growth/secs_per_day
       data%phytos(i)%fT_Method    = pd(list(i))%fT_Method
       data%phytos(i)%theta_growth = pd(list(i))%theta_growth
       data%phytos(i)%T_std        = pd(list(i))%T_std
       data%phytos(i)%T_opt        = pd(list(i))%T_opt
       data%phytos(i)%T_max        = pd(list(i))%T_max
       data%phytos(i)%lightModel   = pd(list(i))%lightModel
       data%phytos(i)%I_K          = pd(list(i))%I_K
       data%phytos(i)%I_S          = pd(list(i))%I_S
       data%phytos(i)%KePHY        = pd(list(i))%KePHY
       data%phytos(i)%f_pr         = pd(list(i))%f_pr
       data%phytos(i)%R_resp       = pd(list(i))%R_resp/secs_per_day
       data%phytos(i)%theta_resp   = pd(list(i))%theta_resp
       data%phytos(i)%k_fres       = pd(list(i))%k_fres
       data%phytos(i)%k_fdom       = pd(list(i))%k_fdom
       data%phytos(i)%salTol       = pd(list(i))%salTol
       data%phytos(i)%S_bep        = pd(list(i))%S_bep
       data%phytos(i)%S_maxsp      = pd(list(i))%S_maxsp
       data%phytos(i)%S_opt        = pd(list(i))%S_opt
       data%phytos(i)%simDINUptake = pd(list(i))%simDINUptake
       data%phytos(i)%simDONUptake = pd(list(i))%simDONUptake
       data%phytos(i)%simNFixation = pd(list(i))%simNFixation
       data%phytos(i)%simINDynamics= pd(list(i))%simINDynamics
       data%phytos(i)%N_o          = pd(list(i))%N_o
       data%phytos(i)%K_N          = pd(list(i))%K_N
       data%phytos(i)%X_ncon       = pd(list(i))%X_ncon
       data%phytos(i)%X_nmin       = pd(list(i))%X_nmin
       data%phytos(i)%X_nmax       = pd(list(i))%X_nmax
       data%phytos(i)%R_nuptake    = pd(list(i))%R_nuptake/secs_per_day
       data%phytos(i)%k_nfix       = pd(list(i))%k_nfix  ! should be between 0 and 1.
       data%phytos(i)%R_nfix       = pd(list(i))%R_nfix/secs_per_day
       data%phytos(i)%simDIPUptake = pd(list(i))%simDIPUptake
       data%phytos(i)%simIPDynamics= pd(list(i))%simIPDynamics
       data%phytos(i)%P_0          = pd(list(i))%P_0
       data%phytos(i)%K_P          = pd(list(i))%K_P
       data%phytos(i)%X_pcon       = pd(list(i))%X_pcon
       data%phytos(i)%X_pmin       = pd(list(i))%X_pmin
       data%phytos(i)%X_pmax       = pd(list(i))%X_pmax
       data%phytos(i)%R_puptake    = pd(list(i))%R_puptake/secs_per_day
       data%phytos(i)%simSiUptake  = pd(list(i))%simSiUptake
       data%phytos(i)%Si_0         = pd(list(i))%Si_0
       data%phytos(i)%K_Si         = pd(list(i))%K_Si
       data%phytos(i)%X_sicon      = pd(list(i))%X_sicon
       ! Hardocoded parameter values
       data%phytos(i)%c1           = 0.1240/60.   ! From Chung et al (2014)
       data%phytos(i)%c3           = 0.0230/60.   !  "
       data%phytos(i)%f1           = 0.675        ! Ross and Sharples (2007)
       data%phytos(i)%f2           = 0.750        !  "
       data%phytos(i)%d_phy        = 1e-5

       ! Register group as a state variable
       data%id_p(i) = aed_define_variable(                                     &
                              TRIM(data%phytos(i)%p_name),                     &
                              'mmol C/m3',                                     &
                              'phytoplankton '//TRIM(data%phytos(i)%p_name),   &
                              pd(list(i))%p_initial,                           &
                              minimum=pd(list(i))%p0,                          &
                              maximum=1e4,                                     &
                              mobility = data%phytos(i)%w_p)

       ! Register rho (density) group as a state variable, if required
       IF (data%phytos(i)%settling == _MOB_STOKES_) THEN
          data%id_rho(i) = aed_define_variable(                                &
                              TRIM(data%phytos(i)%p_name)//'_rho',             &
                              'kg/m3',                                         &
                        'phytoplankton '//TRIM(data%phytos(i)%p_name)//'_rho', &
                              (data%min_rho+data%max_rho)/2.,                  &
                              minimum=data%min_rho,                            &
                              maximum=data%max_rho,                            &
                              mobility = data%phytos(i)%w_p)
       ENDIF

       ! Register internal nitrogen group as a state variable, if required
       IF (data%phytos(i)%simINDynamics /= 0) THEN
          IF(data%phytos(i)%simINDynamics == 1)THEN
            minNut = data%phytos(i)%p0*data%phytos(i)%X_ncon
          ELSE
            minNut = data%phytos(i)%p0*data%phytos(i)%X_nmin
          ENDIF
          ! Register IN group as a state variable
          data%id_in(i) = aed_define_variable(                                 &
                              TRIM(data%phytos(i)%p_name)//'_IN',              &
                              'mmol N/m3',                                     &
                         'phytoplankton '//TRIM(data%phytos(i)%p_name)//'_IN', &
                              pd(list(i))%p_initial*data%phytos(i)%X_ncon,     &
                              minimum=minNut,                                  &
                              mobility = data%phytos(i)%w_p)

       ELSE
         IF (data%phytos(i)%settling == _MOB_MOTILE_) THEN
             data%phytos(i)%settling = _MOB_CONST_
             print *,'Motility can not be simulated as no IN'
         ENDIF
       ENDIF

       ! Register internal phosphorus group as a state variable, if required
       IF (data%phytos(i)%simIPDynamics /= 0) THEN
          IF(data%phytos(i)%simIPDynamics == 1)THEN
            minNut = data%phytos(i)%p0*data%phytos(i)%X_pcon
          ELSE
            minNut = data%phytos(i)%p0*data%phytos(i)%X_pmin
          ENDIF
          ! Register IP group as a state variable
          data%id_ip(i) = aed_define_variable(                                 &
                              TRIM(data%phytos(i)%p_name)//'_IP',              &
                              'mmol P/m3',                                     &
                         'phytoplankton '//TRIM(data%phytos(i)%p_name)//'_IP', &
                              pd(list(i))%p_initial*data%phytos(i)%X_pcon,     &
                              minimum=minNut,                                  &
                              mobility = data%phytos(i)%w_p)
       ENDIF

       ! Group specific diagnostic variables
       data%id_NtoP(i) = aed_define_diag_variable( TRIM(data%phytos(i)%p_name)//'_NtoP','mol N/ mol P', 'internal n:p ratio')
       IF ( diag_level >= 10 ) THEN
          ! Growth controls
          data%id_fI(i)   = aed_define_diag_variable( TRIM(data%phytos(i)%p_name)//'_fI'  , '-', 'fI (0-1)')
          data%id_fNit(i) = aed_define_diag_variable( TRIM(data%phytos(i)%p_name)//'_fNit', '-', 'fNit (0-1)')
          data%id_fPho(i) = aed_define_diag_variable( TRIM(data%phytos(i)%p_name)//'_fPho', '-', 'fPho (0-1)')
          data%id_fSil(i) = aed_define_diag_variable( TRIM(data%phytos(i)%p_name)//'_fSil', '-', 'fSil (0-1)')
          data%id_fT(i)   = aed_define_diag_variable( TRIM(data%phytos(i)%p_name)//'_fT'  , '-', 'fT (>0)')
          data%id_fSal(i) = aed_define_diag_variable( TRIM(data%phytos(i)%p_name)//'_fSal', '-', 'fSal (0-1+)')
          ! Mass flux tracking
          data%id_PhyGPPc(i) = aed_define_diag_variable( TRIM(data%phytos(i)%p_name)//'_gpp_c', &
                                                         'mmol C/m3/d', 'group primary production')
          data%id_PhyRSPc(i) = aed_define_diag_variable( TRIM(data%phytos(i)%p_name)//'_rsp_c', &
                                                         'mmol C/m3/d', 'group respiration')
          data%id_PhyEXCc(i) = aed_define_diag_variable( TRIM(data%phytos(i)%p_name)//'_exc_c', &
                                                         'mmol C/m3/d', 'group excretion/exudation')
          data%id_PhyMORc(i) = aed_define_diag_variable( TRIM(data%phytos(i)%p_name)//'_mor_c', &
                                                         'mmol C/m3/d', 'group mortality')
          data%id_PhySEDc(i) = aed_define_diag_variable( TRIM(data%phytos(i)%p_name)//'_set_c', &
                                                         'mmol C/m3/d', 'group sedimentation')
          data%id_PhyGPPn(i) = aed_define_diag_variable( TRIM(data%phytos(i)%p_name)//'_gpp_n', &
                                                         'mmol N/m3/d', 'group primary production')
          data%id_PhyRSPn(i) = aed_define_diag_variable( TRIM(data%phytos(i)%p_name)//'_rsp_n', &
                                                         'mmol N/m3/d', 'group respiration')
          data%id_PhyEXCn(i) = aed_define_diag_variable( TRIM(data%phytos(i)%p_name)//'_exc_n', &
                                                         'mmol N/m3/d', 'group excretion/exudation')
          data%id_PhyMORn(i) = aed_define_diag_variable( TRIM(data%phytos(i)%p_name)//'_mor_n', &
                                                         'mmol N/m3/d', 'group mortality')
          data%id_PhySEDn(i) = aed_define_diag_variable( TRIM(data%phytos(i)%p_name)//'_set_n', &
                                                         'mmol N/m3/d', 'group sedimentation')
          data%id_PhyGPPp(i) = aed_define_diag_variable( TRIM(data%phytos(i)%p_name)//'_gpp_p', &
                                                         'mmol P/m3/d', 'group primary production')
          data%id_PhyRSPp(i) = aed_define_diag_variable( TRIM(data%phytos(i)%p_name)//'_rsp_p', &
                                                         'mmol P/m3/d', 'group respiration')
          data%id_PhyEXCp(i) = aed_define_diag_variable( TRIM(data%phytos(i)%p_name)//'_exc_p', &
                                                         'mmol P/m3/d', 'group excretion/exudation')
          data%id_PhyMORp(i) = aed_define_diag_variable( TRIM(data%phytos(i)%p_name)//'_mor_p', &
                                                         'mmol P/m3/d', 'group mortality')
          data%id_PhySEDp(i) = aed_define_diag_variable( TRIM(data%phytos(i)%p_name)//'_set_p', &
                                                         'mmol P/m3/d', 'group sedimentation')
          ! Register vertical velocity diagnostic, where relevant
          IF (data%phytos(i)%settling == _MOB_STOKES_ .OR. &
                                   data%phytos(i)%settling == _MOB_MOTILE_) THEN
            data%id_vvel(i) = aed_define_diag_variable( TRIM(data%phytos(i)%p_name)//'_vvel',   &
                                                         'm/d', 'vertical velocity')
          ENDIF
       ENDIF
    ENDDO
END SUBROUTINE aed_phytoplankton_load_params
!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


!###############################################################################
SUBROUTINE aed_define_phytoplankton(data, namlst)
!-------------------------------------------------------------------------------
! Initialise the phytoplankton biogeochemical model
!
!  Here, the aed_p_m namelist is read and te variables exported
!  by the model are registered with AED.
!-------------------------------------------------------------------------------
!ARGUMENTS
   CLASS (aed_phytoplankton_data_t),INTENT(inout) :: data
   INTEGER,INTENT(in) :: namlst
!
!LOCALS
   INTEGER  :: status,i

!  %% NAMELIST
!  %% Last Checked 20/08/2021
   ! Default settings
   INTEGER            :: num_phytos = 0
   INTEGER            :: the_phytos(MAX_PHYTO_TYPES) = 0
   INTEGER            :: settling(MAX_PHYTO_TYPES)     = _MOB_CONST_
   AED_REAL           :: resuspension(MAX_PHYTO_TYPES) = 0.
   CHARACTER(len=64)  :: resus_link='NCS_resus'
   CHARACTER(len=64)  :: phyto_particle_link=''            !   For FV API 2.0 (To be implemented)
   CHARACTER(len=64)  :: p_excretion_target_variable='OGM_dop'
   CHARACTER(len=64)  :: p_mortality_target_variable='OGM_pop'
   CHARACTER(len=64)  :: p1_uptake_target_variable='PHS_frp'
   CHARACTER(len=64)  :: p2_uptake_target_variable=''
   CHARACTER(len=64)  :: n_excretion_target_variable='OGM_don'
   CHARACTER(len=64)  :: n_mortality_target_variable='OGM_pon'
   CHARACTER(len=64)  :: n1_uptake_target_variable='NIT_nit'
   CHARACTER(len=64)  :: n2_uptake_target_variable='NIT_amm'
   CHARACTER(len=64)  :: n3_uptake_target_variable=''
   CHARACTER(len=64)  :: n4_uptake_target_variable=''
   CHARACTER(len=64)  :: c_excretion_target_variable='OGM_doc'
   CHARACTER(len=64)  :: c_mortality_target_variable='OGM_poc'
   CHARACTER(len=64)  :: c_uptake_target_variable=''
   CHARACTER(len=64)  :: do_uptake_target_variable='OXY_oxy'
   CHARACTER(len=64)  :: si_excretion_target_variable=''
   CHARACTER(len=64)  :: si_mortality_target_variable=''
   CHARACTER(len=64)  :: si_uptake_target_variable=''
   CHARACTER(len=128) :: dbase='aed_phyto_pars.nml'
   AED_REAL           :: zerolimitfudgefactor = 0.9 * 3600
   AED_REAL           :: R_mpbg = 0.
   AED_REAL           :: R_mpbr = 0.
   AED_REAL           :: I_Kmpb = 100.
   AED_REAL           :: mpb_max = 1000.
   AED_REAL           :: theta_mpb_growth = 1.05
   AED_REAL           :: theta_mpb_resp   = 1.05
   AED_REAL           :: min_rho = 900.
   AED_REAL           :: max_rho = 1200.
   INTEGER            :: do_mpb = 0
   INTEGER            :: n_zones = 0
   AED_REAL           :: active_zones(1000) = 0

!  From module globals
   LOGICAL  :: extra_debug = .false.      !## Obsolete Use diag_level = 10
!  LOGICAL  :: extra_diag = .false.      !## Obsolete Use diag_level = 10
!  INTEGER  :: diag_level = 10                ! 0 = no diagnostic outputs
!                                             ! 1 = basic diagnostic outputs
!                                             ! 2 = flux rates, and supporitng
!                                             ! 3 = other metrics
!                                             !10 = all debug & checking outputs
!  %% END NAMELIST

   NAMELIST /aed_phytoplankton/ num_phytos, the_phytos, settling, resuspension,&
                    p_excretion_target_variable,p_mortality_target_variable,   &
                     p1_uptake_target_variable, p2_uptake_target_variable,     &
                    n_excretion_target_variable,n_mortality_target_variable,   &
                     n1_uptake_target_variable,n2_uptake_target_variable,      &
                     n3_uptake_target_variable,n4_uptake_target_variable,      &
                    c_excretion_target_variable,c_mortality_target_variable,   &
                      c_uptake_target_variable, do_uptake_target_variable,     &
                    si_excretion_target_variable,si_mortality_target_variable, &
                      si_uptake_target_variable,                               &
                    dbase, zerolimitfudgefactor, extra_debug, extra_diag,      &
                    do_mpb, R_mpbg, R_mpbr, I_Kmpb, mpb_max, min_rho, max_rho, &
                    resus_link, n_zones, active_zones, diag_level,             &
                    theta_mpb_growth,theta_mpb_resp,                           &
                    phyto_particle_link
!-----------------------------------------------------------------------
!BEGIN
   print *,"        aed_phytoplankton configuration"

   ! Read the namelist, and set module parameters
   read(namlst,nml=aed_phytoplankton,iostat=status)
   IF (status /= 0) STOP 'Error reading namelist for &aed_phytoplankton'
   dtlim = zerolimitfudgefactor
   IF( extra_debug ) extra_diag = .true.       ! legacy use of extra_debug
   IF ( extra_diag ) diag_level = 10

   ! Set module parameters
   data%min_rho = min_rho ; data%max_rho = max_rho
   data%do_mpb = do_mpb
   data%R_mpbg = R_mpbg/secs_per_day   ; data%R_mpbr = R_mpbr/secs_per_day
   data%I_Kmpb = I_Kmpb   ; data%mpb_max = mpb_max
   data%theta_mpb_growth = theta_mpb_growth   ; data%theta_mpb_resp = theta_mpb_resp
   ALLOCATE(data%resuspension(num_phytos)); data%resuspension = resuspension(1:num_phytos)
   data%n_zones = n_zones
   IF( n_zones>0 ) THEN
     ALLOCATE(data%active_zones(n_zones))
     DO i=1,n_zones
       data%active_zones(i) = active_zones(i)
     ENDDO
   ENDIF

   ! Store species parameter values in our own derived type, and register species
   CALL aed_phytoplankton_load_params(data,dbase,num_phytos,the_phytos,settling,resuspension)

   CALL aed_bio_temp_function( data%num_phytos,                                &
                               data%phytos%theta_growth,                       &
                               data%phytos%T_std,                              &
                               data%phytos%T_opt,                              &
                               data%phytos%T_max,                              &
                               data%phytos%aTn,                                &
                               data%phytos%bTn,                                &
                               data%phytos%kTn,                                &
                               data%phytos%p_name)

   ! Register microphytbenthos as a state variable
   IF (data%do_mpb>0) THEN
     data%id_mpb = aed_define_sheet_variable(  'mpb',                          &
                                               'mmol C/m2',                    &
                                               'microphytobenthos biomass',    &
                                                0.001,                         &
                                                minimum=0.001)
   ENDIF

   ! Register link to nutrient pools, if variable names are provided in namelist
   data%do_Pexc = p_excretion_target_variable .NE. ''
   IF (data%do_Pexc) THEN
     data%id_Pexctarget = aed_locate_variable(p_excretion_target_variable)
   ENDIF
   data%do_Nexc = n_excretion_target_variable .NE. ''
   IF (data%do_Nexc) THEN
     data%id_Nexctarget = aed_locate_variable(n_excretion_target_variable)
   ENDIF
   data%do_Cexc = c_excretion_target_variable .NE. ''
   IF (data%do_Cexc) THEN
     data%id_Cexctarget = aed_locate_variable(c_excretion_target_variable)
   ENDIF
   data%do_Siexc = si_excretion_target_variable .NE. ''
   IF (data%do_Siexc) THEN
     data%id_Siexctarget = aed_locate_variable(si_excretion_target_variable)
   ENDIF

   data%do_Pmort = p_mortality_target_variable .NE. ''
   IF (data%do_Pmort) THEN
     data%id_Pmorttarget = aed_locate_variable(p_mortality_target_variable)
   ENDIF
   data%do_Nmort = n_mortality_target_variable .NE. ''
   IF (data%do_Nmort) THEN
     data%id_Nmorttarget = aed_locate_variable(n_mortality_target_variable)
   ENDIF
   data%do_Cmort = c_mortality_target_variable .NE. ''
   IF (data%do_Cmort) THEN
     data%id_Cmorttarget = aed_locate_variable(c_mortality_target_variable)
   ENDIF
   data%do_Simort = si_mortality_target_variable .NE. ''
   IF (data%do_Simort) THEN
     data%id_Simorttarget = aed_locate_variable(si_mortality_target_variable)
   ENDIF

   data%npup = 0
   IF (p1_uptake_target_variable .NE. '') data%npup = 1
   IF (p2_uptake_target_variable .NE. '') data%npup = 2
   data%do_Puptake = .FALSE.
   IF (data%npup>0) data%do_Puptake=.TRUE.
   IF (data%do_Puptake) THEN
   ! IF (data%npup>0) THEN ; data%id_Pupttarget(1) = aed_locate_variable(p1_uptake_target_variable); ifrp=1 ; ENDIF
   ! IF (data%npup>1) THEN ; data%id_Pupttarget(2) = aed_locate_variable(p2_uptake_target_variable); idop=2 ; ENDIF
     !# CAB ifrp and idop are now constants in bio_utils
     IF (data%npup>0) data%id_Pupttarget(ifrp) = aed_locate_variable(p1_uptake_target_variable)
     IF (data%npup>1) data%id_Pupttarget(idop) = aed_locate_variable(p2_uptake_target_variable)
   ENDIF
   data%nnup = 0
   IF (n1_uptake_target_variable .NE. '') data%nnup = 1
   IF (n2_uptake_target_variable .NE. '') data%nnup = 2
   IF (n3_uptake_target_variable .NE. '') data%nnup = 3
   IF (n4_uptake_target_variable .NE. '') data%nnup = 4
   data%do_Nuptake = .false.
   IF (data%nnup>0) data%do_Nuptake=.true.
   IF (data%do_Nuptake) THEN
   ! IF (data%nnup>0) THEN ; data%id_Nupttarget(1) = aed_locate_variable( n1_uptake_target_variable); ino3=1 ; ENDIF
   ! IF (data%nnup>1) THEN ; data%id_Nupttarget(2) = aed_locate_variable( n2_uptake_target_variable); inh4=2 ; ENDIF
   ! IF (data%nnup>2) THEN ; data%id_Nupttarget(3) = aed_locate_variable( n3_uptake_target_variable); idon=3 ; ENDIF
   ! IF (data%nnup>3) THEN ; data%id_Nupttarget(4) = aed_locate_variable( n4_uptake_target_variable); in2 =4 ; ENDIF
     !# CAB ino3, inh4, idon and in2 are now constants in bio_utils
     IF (data%nnup>0) data%id_Nupttarget(ino3) = aed_locate_variable( n1_uptake_target_variable)
     IF (data%nnup>1) data%id_Nupttarget(inh4) = aed_locate_variable( n2_uptake_target_variable)
     IF (data%nnup>2) data%id_Nupttarget(idon) = aed_locate_variable( n3_uptake_target_variable)
     IF (data%nnup>3) data%id_Nupttarget(in2)  = aed_locate_variable( n4_uptake_target_variable)
   ENDIF
   data%do_Cuptake = c_uptake_target_variable .NE. ''
   IF (data%do_Cuptake) THEN
     data%id_Cupttarget = aed_locate_variable( c_uptake_target_variable)
   ENDIF
   data%do_DOuptake = do_uptake_target_variable .NE. ''
   IF (data%do_DOuptake) THEN
     data%id_DOupttarget = aed_locate_variable( do_uptake_target_variable)
   ENDIF
   data%do_Siuptake = si_uptake_target_variable .NE. ''
   IF (data%do_Siuptake) THEN
     data%id_Siupttarget = aed_locate_variable( si_uptake_target_variable)
   ENDIF

   !-- Create sedimentation and resuspension diagnostic variables
   data%id_Psed_phy = aed_define_diag_variable('set','mmol C/m2/d','phytoplankton sedimentation')
  !data%id_Psed_phy = aed_define_diag_variable('Psed_phy','mmol C/m2/d','PHY sedimentation')
   data%id_d_res = aed_define_sheet_diag_variable('res','mmol C/m2/d', 'phytoplankton resuspension')

   !-- Check for resuspension link variable
   IF ( .NOT.resus_link .EQ. '' ) THEN
      data%id_l_resus   = aed_locate_sheet_variable(TRIM(resus_link))
   ELSE
      data%id_l_resus   = 0
      data%resuspension = 0.
   ENDIF

   ! Register general/community level diagnostic variables
   data%id_chl = aed_define_diag_variable('tchla'  ,'ug/L', 'total chlorophyll-a')
   data%id_TIN = aed_define_diag_variable('in'     ,'mmol N/m3', 'total phytoplankton nitrogen')
   data%id_TIP = aed_define_diag_variable('ip'     ,'mmol P/m3', 'total phytoplankton phosphorus')

   data%id_GPP = aed_define_diag_variable('gpp'    ,'mmol C/m3/d',  'gross primary production')
   data%id_NCP = aed_define_diag_variable('ncp'    ,'mmol C/m3/d',  'net community production')

   data%id_NUP1= aed_define_diag_variable('upt_no3','mmol N/m3/d','nitrogen (NO3) uptake')
   data%id_NUP2= aed_define_diag_variable('upt_nh4','mmol N/m3/d','nitrogen (NH4) uptake')
   data%id_NUP4= aed_define_diag_variable('upt_n2' ,'mmol N/m3/d','nitrogen (N2) uptake')
   data%id_PUP = aed_define_diag_variable('upt_po4','mmol P/m3/d','phosphate uptake')
   data%id_CUP = aed_define_diag_variable('upt_dic','mmol C/m3/d','carbon uptake')

   IF ( diag_level >= 10 ) data%id_dPAR = aed_define_diag_variable('par'  ,'W/m2'   ,'photosynthetically active radiation')
   IF ( diag_level >= 10 ) data%id_TPHY = aed_define_diag_variable('tphy' ,'mmol/m3','total phytoplankton')
   IF ( diag_level >= 10 ) data%id_PPR  = aed_define_diag_variable('ppr'  ,'-'      ,'phytoplankton p/r ratio (gross)')
   IF ( diag_level >= 10 ) data%id_NPR  = aed_define_diag_variable('npr'  ,'-'      ,'phytoplankton p/r ratio (net)')

   ! Register benthic phyto diagnostic variables
   IF(do_mpb>0) THEN
     data%id_d_MPB = aed_define_sheet_diag_variable('mpb_ben','mmol C/m2'  ,'microphytobenthos density')
     data%id_d_BPP = aed_define_sheet_diag_variable('gpp_ben','mmol C/m2/d','benthic gross productivity')
     data%id_d_BCP = aed_define_sheet_diag_variable('ncp_ben','mmol C/m2/d','benthic net productivity')
     data%id_d_mpbv= aed_define_sheet_diag_variable('swi_mpb','mmol C/m2/d','mpb vertical exchange')
   ENDIF

   ! Register environmental dependencies
   data%id_sedzone = aed_locate_sheet_global('sed_zone')
   data%id_I_0     = aed_locate_sheet_global('par_sf')
   data%id_tem     = aed_locate_global('temperature')
   data%id_sal     = aed_locate_global('salinity')
   data%id_extc    = aed_locate_global('extc_coef')
   data%id_dz      = aed_locate_global('layer_ht')
   data%id_dens    = aed_locate_global('density')
   data%id_par     = aed_locate_global('par')

END SUBROUTINE aed_define_phytoplankton
!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


!###############################################################################
SUBROUTINE aed_calculate_phytoplankton(data,column,layer_idx)
!------------------------------------------------------------------------------+
! Right hand sides of phytoplankton biogeochemical model
!------------------------------------------------------------------------------+
!ARGUMENTS
   CLASS (aed_phytoplankton_data_t),INTENT(in) :: data
   TYPE (aed_column_t),INTENT(inout) :: column(:)
   INTEGER,INTENT(in) :: layer_idx
!
!LOCALS
   AED_REAL :: phy, tphy, tin, tip, tchla
   AED_REAL :: INi, IPi
   AED_REAL :: pup
   AED_REAL :: no3up,nh4up
   AED_REAL :: cup, rsiup
   AED_REAL :: temp, par, Io, salinity, extc, dz
   AED_REAL :: primprod(data%num_phytos), exudation(data%num_phytos), &
               a_nfix(data%num_phytos), respiration(data%num_phytos)
   AED_REAL :: cuptake(data%num_phytos), cexcretion(data%num_phytos), cmortality(data%num_phytos)
   AED_REAL :: nuptake(data%num_phytos,1:4), nexcretion(data%num_phytos), nmortality(data%num_phytos)
   AED_REAL :: puptake(data%num_phytos,1:2), pexcretion(data%num_phytos), pmortality(data%num_phytos)
   AED_REAL :: siuptake(data%num_phytos), siexcretion(data%num_phytos), simortality(data%num_phytos)
   AED_REAL :: fT, fNit, fPho, fSil, fI, fXl, fSal, PNf
   AED_REAL :: upTot,net_cuptake

   INTEGER  :: phy_i,c
   AED_REAL :: flux, available

!------------------------------------------------------------------------------+
!BEGIN
   ! Retrieve current environmental conditions.
   temp = _STATE_VAR_(data%id_tem)      ! local temperature
   salinity = _STATE_VAR_(data%id_sal)  ! local salinity
   par = _STATE_VAR_(data%id_par)       ! local photosynth. active radiation
   Io = _STATE_VAR_S_(data%id_I_0)      ! surface short wave radiation

   ! Retrieve current (local) state variable values.
   pup = 0.
   IF (data%do_Puptake)  pup = _STATE_VAR_(data%id_Pupttarget(1))

   no3up = 0.
   nh4up = 0.
   IF (data%do_Nuptake) THEN
       no3up = _STATE_VAR_(data%id_Nupttarget(1))
       nh4up = _STATE_VAR_(data%id_Nupttarget(2))
   ENDIF
   cup = 0.
   IF (data%do_Cuptake)  cup = _STATE_VAR_(data%id_Cupttarget)
   rsiup = 0.
   IF (data%do_Siuptake)  rsiup = _STATE_VAR_(data%id_Siupttarget)

   tphy = 0.0
   tchla = 0.0
   tin  = 0.0
   tip  = 0.0

   INi = 0.
   IPi = 0.

   !---------------------------------------------------------------------------+
   DO phy_i=1,data%num_phytos

      primprod(phy_i)    = zero_
      exudation(phy_i)   = zero_
      a_nfix(phy_i)      = zero_
      respiration(phy_i) = zero_

      cuptake(phy_i)     = zero_
      cexcretion(phy_i)  = zero_
      cmortality(phy_i)  = zero_
      nuptake(phy_i,:)   = zero_
      nexcretion(phy_i)  = zero_
      nmortality(phy_i)  = zero_
      puptake(phy_i,:)   = zero_
      pexcretion(phy_i)  = zero_
      pmortality(phy_i)  = zero_

      ! Retrieve this phytoplankton group
      phy = _STATE_VAR_(data%id_p(phy_i))

      !------------------------------------------------------------------------+
      ! Get the temperature limitation function
      fT = fTemp_function(data%phytos(phy_i)%fT_Method,    &
                          data%phytos(phy_i)%T_max,        &
                          data%phytos(phy_i)%T_std,        &
                          data%phytos(phy_i)%theta_growth, &
                          data%phytos(phy_i)%aTn,          &
                          data%phytos(phy_i)%bTn,          &
                          data%phytos(phy_i)%kTn,temp)

      !------------------------------------------------------------------------+
      ! Get the light and nutrient limitation.

      ! NITROGEN.
      fNit = 0.0
      IF(data%phytos(phy_i)%simINDynamics /= 0) THEN
         ! IN variable available
         INi = _STATE_VAR_(data%id_in(phy_i))
      ELSE
         ! Assumed constant IN:
         INi = phy*data%phytos(phy_i)%X_ncon
      END IF

      ! Estimate fN limitation from IN or ext N value
      IF(data%phytos(phy_i)%simINDynamics > 1) THEN
         IF (phy > data%phytos(phy_i)%p0) THEN
            fNit = INi / phy
            fNit = phyto_fN(data%phytos,phy_i,IN=fNit)
         ENDIF
         IF (phy > zero_ .AND. phy <= data%phytos(phy_i)%p0) THEN
            fNit = phyto_fN(data%phytos,phy_i,din=no3up+nh4up)
         ENDIF
      ELSE
         fNit = phyto_fN(data%phytos,phy_i,din=no3up+nh4up)
      ENDIF
      IF (data%phytos(phy_i)%simNFixation /= 0) THEN
         ! Nitrogen fixer: apply no N limitation. N Fixation ability
         ! depends on DIN concentration
         a_nfix(phy_i) = (one_ - fNit)
         fNit = one_
      ENDIF


      ! PHOSPHOROUS.
      fPho = zero_
      IF (data%phytos(phy_i)%simIPDynamics /= 0) THEN
         ! IP variable available
         IPi = _STATE_VAR_(data%id_ip(phy_i))
      ELSE
         ! Assumed constant IP:
         IPi = phy*data%phytos(phy_i)%X_pcon
      END IF

      ! Estimate fP limitation from IP or ext P value
      IF (data%phytos(phy_i)%simIPDynamics > 1) THEN
         IF (phy > data%phytos(phy_i)%p0) THEN
            fPho = IPi / phy
            fPho = phyto_fP(data%phytos,phy_i,IP=fPho)
         ENDIF
         IF (phy > zero_ .AND. phy <= data%phytos(phy_i)%p0) THEN
            fPho = phyto_fP(data%phytos,phy_i,frp=pup)
         ENDIF
      ELSE
         fPho = phyto_fP(data%phytos,phy_i,frp=pup)
      ENDIF

      ! SILICA.
      fSil = phyto_fSi(data%phytos,phy_i,rsiup)


      ! LIGHT
      extc = _STATE_VAR_(data%id_extc)
      ! dz = 0.5     !MH: to fix
      dz = _STATE_VAR_(data%id_dz)
      fI = photosynthesis_irradiance(data%phytos(phy_i)%lightModel, &
               data%phytos(phy_i)%I_K, data%phytos(phy_i)%I_S, par, extc, Io, dz)
      ! fI = 0.1


      ! METAL AND TOXIC EFFECTS
      fXl = 1.0

      !------------------------------------------------------------------------+
      ! Primary production rate
      primprod(phy_i) = data%phytos(phy_i)%R_growth * fT * findMin(fI,fNit,fPho,fSil) * fxl

      ! Adjust primary production rate for nitrogen fixers
      IF (data%phytos(phy_i)%simNFixation /= 0) THEN
         ! Nitrogen fixing species, and the growth rate to  must be reduced
         ! to compensate for the increased metabolic cost of this process
         primprod(phy_i) = primprod(phy_i) * (data%phytos(phy_i)%k_nfix + &
                           (1.0-a_nfix(phy_i))*(1.0-data%phytos(phy_i)%k_nfix))
      ENDIF

      !------------------------------------------------------------------------+
      ! Respiration and general metabolic loss
      respiration(phy_i) = bio_respiration(data%phytos(phy_i)%R_resp,data%phytos(phy_i)%theta_resp,temp)

      ! Salinity stress effect on respiration or growth
      fSal =  phyto_salinity(data%phytos,phy_i,salinity)
      IF( data%phytos(phy_i)%salTol < 0) THEN
        ! salTol is set for growth supression
        primprod(phy_i) = primprod(phy_i) * fSal
      ELSE
        ! salTol is set for respiration/mortality enhancement
        respiration(phy_i) = respiration(phy_i) * fSal
      ENDIF

      ! photo-exudation
      exudation(phy_i) = primprod(phy_i)*data%phytos(phy_i)%f_pr

      ! Limit respiration if at the min biomass to prevent
      ! leak in the C mass balance
      IF (phy <= data%phytos(phy_i)%p0) THEN
         respiration(phy_i) = zero_
         exudation(phy_i) = zero_
      ENDIF

      ! write(*,"(4X,'limitations (fT,fI,fN,fP,fSi,Io, par, mu): ',9F9.2)")fT,fI,fNit,fPho,fSil,Io,par,primprod*secs_per_day

      !------------------------------------------------------------------------+
      ! Carbon uptake and excretion
      cuptake(phy_i)    = -primprod(phy_i) * phy
      cexcretion(phy_i) = (data%phytos(phy_i)%k_fdom*(1.0-data%phytos(phy_i)%k_fres)*respiration(phy_i)+exudation(phy_i)) * phy
      cmortality(phy_i) = ((1.0-data%phytos(phy_i)%k_fdom)*(1.0-data%phytos(phy_i)%k_fres)*respiration(phy_i)) * phy

      ! Nitrogen uptake and excretion
      CALL phyto_internal_nitrogen(data%phytos,phy_i,data%do_N2uptake,phy,INi,primprod(phy_i),&
                             fT,no3up,nh4up,a_nfix(phy_i),respiration(phy_i),exudation(phy_i),PNf,&
                                   nuptake(phy_i,:),nexcretion(phy_i),nmortality(phy_i))

      ! Phosphorus uptake and excretion
      CALL phyto_internal_phosphorus(data%phytos,phy_i,data%npup,phy,IPi,primprod(phy_i),&
                                 fT,pup,respiration(phy_i),exudation(phy_i),&
                                         puptake(phy_i,:),pexcretion(phy_i),pmortality(phy_i))

      ! Silica uptake and excretion
      IF (data%phytos(phy_i)%simSiUptake > 0) THEN
         siuptake(phy_i)    =-data%phytos(phy_i)%X_sicon * primprod(phy_i) * phy
         siexcretion(phy_i) = data%phytos(phy_i)%X_sicon * (data%phytos(phy_i)%k_fdom*respiration(phy_i)+exudation(phy_i)) * phy
         simortality(phy_i) = data%phytos(phy_i)%X_sicon * ((1.0-data%phytos(phy_i)%k_fdom)*respiration(phy_i)) * phy
      ELSE
         siuptake(phy_i)    = zero_
         siexcretion(phy_i) = zero_
         simortality(phy_i) = zero_
      ENDIF

      !------------------------------------------------------------------------+
      ! Diagnostic info
      _DIAG_VAR_(data%id_NtoP(phy_i)) =  INi/IPi

      IF ( diag_level >= 10 ) THEN
         ! Growth controls
         _DIAG_VAR_(data%id_fT(phy_i))   =  fT
         _DIAG_VAR_(data%id_fI(phy_i))   =  fI
         _DIAG_VAR_(data%id_fNit(phy_i)) =  fNit
         _DIAG_VAR_(data%id_fPho(phy_i)) =  fPho
         _DIAG_VAR_(data%id_fSil(phy_i)) =  fSil
         _DIAG_VAR_(data%id_fSal(phy_i)) =  fSal
      ENDIF
   END DO


   !---------------------------------------------------------------------------+
   ! Check uptake values for availability to prevent -ve numbers, and report

   ! pup   - p available
   ! no3up - no3 available
   ! nh4up - nh4 available
   ! cup   - c available
   ! rsiup - Si available

   IF (data%do_Puptake) THEN
      upTot = sum(puptake(:,1))*dtlim
      IF ( abs(upTot) > 1.e-10 .and. upTot >= pup ) THEN
         DO phy_i=1,data%num_phytos
            puptake(phy_i,1) = (pup*0.99/dtlim) * (puptake(phy_i,1)/upTot)
         ENDDO
      ENDIF
   ENDIF

   IF (data%do_Nuptake) THEN
      upTot = sum(nuptake(:,1))*dtlim
      IF ( abs(upTot) > 1.e-10 .and. upTot >= no3up ) THEN
         DO phy_i=1,data%num_phytos
            nuptake(phy_i,1) = (no3up*0.99/dtlim) * (nuptake(phy_i,1)/upTot)
         ENDDO
      ENDIF

      upTot = sum(nuptake(:,2))*dtlim
      IF ( abs(upTot) > 1.e-10 .and. upTot >= nh4up ) THEN
         DO phy_i=1,data%num_phytos
            nuptake(phy_i,2) = (nh4up*0.99/dtlim) * (nuptake(phy_i,2)/upTot)
         ENDDO
      ENDIF
   ENDIF
   IF (data%do_Cuptake) THEN
      upTot = sum(cuptake)*dtlim
      IF ( abs(upTot) > 1.e-10 .and. upTot >= cup ) THEN
         DO phy_i=1,data%num_phytos
            cuptake(phy_i) = (cup*0.99/dtlim) * (cuptake(phy_i)/upTot)
         ENDDO
      ENDIF
   ENDIF
!  IF (data%do_DOuptake) THEN
!     !
!  ENDIF
   IF (data%do_Siuptake) THEN
      upTot = sum(siuptake)*dtlim
      IF ( abs(upTot) > 1.e-10 .and. upTot >= rsiup ) THEN
         DO phy_i=1,data%num_phytos
            siuptake(phy_i) = (rsiup*0.99/dtlim) * (siuptake(phy_i)/upTot)
         ENDDO
      ENDIF
   ENDIF

   DO phy_i=1,data%num_phytos
     IF ( diag_level >= 10 ) THEN
          phy = _STATE_VAR_(data%id_p(phy_i))
          IF (data%phytos(phy_i)%simINDynamics /= 0) THEN; INi = _STATE_VAR_(data%id_in(phy_i))
          ELSE; INi = phy*data%phytos(phy_i)%X_ncon; END IF
          IF (data%phytos(phy_i)%simIPDynamics /= 0) THEN; IPi = _STATE_VAR_(data%id_ip(phy_i))
          ELSE; IPi = phy*data%phytos(phy_i)%X_pcon; END IF
          ! Carbon fluxes
          _DIAG_VAR_(data%id_PhyGPPc(phy_i)) =  cuptake(phy_i) * secs_per_day
          _DIAG_VAR_(data%id_PhyRSPc(phy_i)) = ( -respiration(phy_i)*data%phytos(phy_i)%k_fres*phy) * secs_per_day
          _DIAG_VAR_(data%id_PhyEXCc(phy_i)) = cexcretion(phy_i) * secs_per_day
          _DIAG_VAR_(data%id_PhyMORc(phy_i)) = cmortality(phy_i) * secs_per_day
          ! Nitrogen fluxes
          _DIAG_VAR_(data%id_PhyGPPn(phy_i)) =  sum(nuptake(phy_i,:)) * secs_per_day
          _DIAG_VAR_(data%id_PhyRSPn(phy_i)) = ( -respiration(phy_i)*data%phytos(phy_i)%k_fres*INi) * secs_per_day
          _DIAG_VAR_(data%id_PhyEXCn(phy_i)) = nexcretion(phy_i) * secs_per_day
          _DIAG_VAR_(data%id_PhyMORn(phy_i)) = nmortality(phy_i) * secs_per_day
          ! Phosphorus fluxes
          _DIAG_VAR_(data%id_PhyGPPp(phy_i)) =  puptake(phy_i,1) * secs_per_day
          _DIAG_VAR_(data%id_PhyRSPp(phy_i)) = ( -respiration(phy_i)*data%phytos(phy_i)%k_fres*IPi) * secs_per_day
          _DIAG_VAR_(data%id_PhyEXCp(phy_i)) = pexcretion(phy_i) * secs_per_day
          _DIAG_VAR_(data%id_PhyMORp(phy_i)) = pmortality(phy_i) * secs_per_day
     ENDIF
   ENDDO
   !---------------------------------------------------------------------------+
   ! SET TEMPORAL DERIVATIVES FOR ODE SOLVER
   net_cuptake = zero_
   DO phy_i=1,data%num_phytos

      !------------------------------------------------------------------------+
      !# PHYTOPLANKTON PRODUCTION & RESPIRATION
      phy = _STATE_VAR_(data%id_p(phy_i))
      flux = (primprod(phy_i) - respiration(phy_i) - exudation(phy_i)) * phy
      available = MAX(zero_, phy - data%phytos(phy_i)%p0)
      IF ( -flux*dtlim > available  ) flux = -0.99*available/dtlim
      _FLUX_VAR_(data%id_p(phy_i)) = _FLUX_VAR_(data%id_p(phy_i)) + (flux)

      !------------------------------------------------------------------------+
      !# PHYTOPLANKTON INTERNAL NITROGEN
      IF (data%phytos(phy_i)%simINDynamics /= 0) THEN
         ! _FLUX_VAR_(data%id_in(phy_i)) = _FLUX_VAR_(data%id_in(phy_i)) +     &
         !      ( (-sum(nuptake) - nexcretion(phy_i) - nmortality(phy_i) )*INi )
         INi = _STATE_VAR_(data%id_in(phy_i))
         flux = (-sum(nuptake(phy_i,:)) - nexcretion(phy_i) - nmortality(phy_i))
         available = MAX(zero_, INi - data%phytos(phy_i)%X_nmin*phy)
         IF ( -flux*dtlim > available  ) flux = -0.99*available/dtlim
         _FLUX_VAR_(data%id_in(phy_i)) = _FLUX_VAR_(data%id_in(phy_i)) + (flux)
      ELSE
         ! Assumed constant IN
         INi = phy*data%phytos(phy_i)%X_ncon
      ENDIF

      !------------------------------------------------------------------------+
      !# PHYTOPLANKTON INTERNAL PHOSPHORUS
      IF (data%phytos(phy_i)%simIPDynamics /= 0) THEN
         ! _FLUX_VAR_(data%id_ip(phy_i)) = _FLUX_VAR_(data%id_ip(phy_i)) +     &
         !  ( (-sum(puptake(phy_i,:)) - pexcretion(phy_i) - pmortality(phy_i)) )
         IPi = _STATE_VAR_(data%id_ip(phy_i))
         flux = (-sum(puptake(phy_i,:)) - pexcretion(phy_i) - pmortality(phy_i))
         available = MAX(zero_, IPi - data%phytos(phy_i)%X_pmin*phy)
         IF ( -flux*dtlim > available  ) flux = -0.99*available/dtlim
         _FLUX_VAR_(data%id_ip(phy_i)) = _FLUX_VAR_(data%id_ip(phy_i)) + (flux)
      ELSE
         ! Assumed constant IP
         IPi = phy*data%phytos(phy_i)%X_pcon
      ENDIF

      !------------------------------------------------------------------------+
      !# PHYTOPLANKTON CELL DENSITY
      IF ( data%id_rho(phy_i)>0 ) THEN
         ! density increases during carbohydrate creation (daytime)
         flux = zero_
         IF( par>zero_ ) THEN
           flux = data%phytos(phy_i)%c1 * &
              (one_ - EXP(-par/data%phytos(phy_i)%I_K) ) - data%phytos(phy_i)%c3
         ELSE
           ! darkness
           flux = -data%phytos(phy_i)%c3
         ENDIF
        _FLUX_VAR_(data%id_rho(phy_i)) = _FLUX_VAR_(data%id_rho(phy_i)) + flux
         ! check maximum/minimum density are not exceeded
         IF( _STATE_VAR_(data%id_rho(phy_i))>data%max_rho ) THEN
            _FLUX_VAR_(data%id_rho(phy_i)) =zero_
            _STATE_VAR_(data%id_rho(phy_i))=data%max_rho
         ENDIF
         IF( _STATE_VAR_(data%id_rho(phy_i))<data%min_rho ) THEN
             _FLUX_VAR_(data%id_rho(phy_i)) =zero_
             _STATE_VAR_(data%id_rho(phy_i))=data%min_rho
         ENDIF
      ENDIF

      !------------------------------------------------------------------------+
      ! BIOGEOCHEMICAL FEEDBACKS
      ! Now manage uptake of nutrients, CO2 and DO - these cumulative fluxes
      ! are already limited above loop
      IF (data%do_Puptake) THEN
         DO c = 1,data%npup
            _FLUX_VAR_(data%id_Pupttarget(c)) =     &
                              _FLUX_VAR_(data%id_Pupttarget(c)) + ( puptake(phy_i,c))
         ENDDO
      ENDIF
      IF (data%do_Nuptake) THEN
         DO c = 1,data%nnup
            _FLUX_VAR_(data%id_Nupttarget(c)) =     &
                               _FLUX_VAR_(data%id_Nupttarget(c)) + ( nuptake(phy_i,c))
         ENDDO
      ENDIF
      IF (data%do_Cuptake) THEN
         _FLUX_VAR_(data%id_Cupttarget) = _FLUX_VAR_(data%id_Cupttarget) +           &
                       (  cuptake(phy_i) + respiration(phy_i)*data%phytos(phy_i)%k_fres*phy )
      ENDIF
      net_cuptake = net_cuptake + (-cuptake(phy_i) - respiration(phy_i)*data%phytos(phy_i)%k_fres*phy)
      IF (data%do_DOuptake) THEN
         _FLUX_VAR_(data%id_DOupttarget) = _FLUX_VAR_(data%id_DOupttarget) +         &
                   ( -cuptake(phy_i) - respiration(phy_i)*data%phytos(phy_i)%k_fres*phy )
      ENDIF
      IF (data%do_Siuptake) THEN
         _FLUX_VAR_(data%id_Siupttarget) = _FLUX_VAR_(data%id_Siupttarget) + ( siuptake(phy_i))
      ENDIF
      ! Now manage mortality contributions to POM
      IF (data%do_Pmort) THEN
         _FLUX_VAR_(data%id_Pmorttarget) = _FLUX_VAR_(data%id_Pmorttarget) + (pmortality(phy_i))
      ENDIF
      IF (data%do_Nmort) THEN
         _FLUX_VAR_(data%id_Nmorttarget) = _FLUX_VAR_(data%id_Nmorttarget) + (nmortality(phy_i))
      ENDIF
      IF (data%do_Cmort) THEN
         _FLUX_VAR_(data%id_Cmorttarget) = _FLUX_VAR_(data%id_Cmorttarget) + (cmortality(phy_i))
      ENDIF
      IF (data%do_Simort) THEN
         _FLUX_VAR_(data%id_Simorttarget) = _FLUX_VAR_(data%id_Simorttarget) + (simortality(phy_i))
      ENDIF
      ! Now manage excretion/exudation contributions to DOM
      IF (data%do_Pexc) THEN
         _FLUX_VAR_(data%id_Pexctarget) = _FLUX_VAR_(data%id_Pexctarget) + (pexcretion(phy_i))
      ENDIF
      IF (data%do_Nexc) THEN
         _FLUX_VAR_(data%id_Nexctarget) = _FLUX_VAR_(data%id_Nexctarget) + (nexcretion(phy_i))
      ENDIF
      IF (data%do_Cexc) THEN
         _FLUX_VAR_(data%id_Cexctarget) = _FLUX_VAR_(data%id_Cexctarget) + (cexcretion(phy_i))
      ENDIF
      IF (data%do_Siexc) THEN
         _FLUX_VAR_(data%id_Siexctarget) = _FLUX_VAR_(data%id_Siexctarget) + (siexcretion(phy_i))
      ENDIF

      !------------------------------------------------------------------------+
      ! UPDATE DIAGNOSTIC VARIABLES

      ! total phytoplankton carbon
      tphy = tphy + phy

      ! total chlorophyll-a
      IF (data%phytos(phy_i)%Xcc > 0.1) THEN
        !-- assume Xcc (mol C/ mol chla) is a constant
        tchla = tchla + ( phy / data%phytos(phy_i)%Xcc ) * 12.0
      ELSE
        !-- use dynamic equation (Eq 13: of Baklouti, Cloern et al. 1995)
        !-- theta = 1/Xcc [mg Chl (mg C)1] = 0.003 + 0.0154  e^0.050T  e^0.059E mu
        tchla = tchla + ( phy * (0.003 + 0.0154 * exp(0.050*temp) * exp(0.059*par) &
                        * primprod(phy_i)))
      ENDIF

      ! total internal nutrients
      tin = tin + INi
      tip = tip + IPi
   ENDDO

   !---------------------------------------------------------------------------+
   ! Set diagnostic arrays for combined assemblage properties
   _DIAG_VAR_(data%id_GPP) =  sum(-cuptake)*secs_per_day
   _DIAG_VAR_(data%id_NCP) =  net_cuptake*secs_per_day
   IF ( diag_level >= 10 ) _DIAG_VAR_(data%id_PPR) =  -999. !sum(cuptake) / ( sum(cuptake) - net_cuptake)
   IF ( diag_level >= 10 ) _DIAG_VAR_(data%id_NPR) =  -999. !net_cuptake / ( sum(cuptake) - net_cuptake)
   _DIAG_VAR_(data%id_NUP1)=  sum(-nuptake(:,1))*secs_per_day
   _DIAG_VAR_(data%id_NUP2)=  sum(-nuptake(:,2))*secs_per_day
   _DIAG_VAR_(data%id_NUP4)=  sum(-nuptake(:,in2))*secs_per_day
   _DIAG_VAR_(data%id_PUP) =  sum(-puptake)*secs_per_day
   _DIAG_VAR_(data%id_CUP) =  sum(-cuptake)*secs_per_day

   IF ( diag_level >= 10 ) _DIAG_VAR_(data%id_dPAR) =  par
   _DIAG_VAR_(data%id_chl)=  tchla
   IF ( diag_level >= 10 ) _DIAG_VAR_(data%id_TPHY) =  tphy
   _DIAG_VAR_(data%id_TIN)  =  tin
   _DIAG_VAR_(data%id_TIP)  =  tip

END SUBROUTINE aed_calculate_phytoplankton
!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


!###############################################################################
SUBROUTINE aed_calculate_benthic_phytoplankton(data,column,layer_idx)
!------------------------------------------------------------------------------+
! Calculate sedimentation of phytoplankton.
! Everything in units per surface area (not volume!) per time.
!------------------------------------------------------------------------------+
!ARGUMENTS
   CLASS (aed_phytoplankton_data_t),INTENT(in) :: data
   TYPE (aed_column_t),INTENT(inout) :: column(:)
   INTEGER,INTENT(in) :: layer_idx
!
!LOCALS
   INTEGER  :: phy_i
   AED_REAL :: phy,mpb,temp,extc,par,dz,Io,fI,matz        ! State
   AED_REAL :: Fsed_phy,Psed_phy,mpb_flux,mpb_prod,mpb_resp,resus,phy_sed_frac
!
!------------------------------------------------------------------------------+
!BEGIN

   !-- Process microphytobenthos (MPB) - sedimented phytoplankton
   IF ( data%do_mpb>0 ) THEN
     ! Get local conditions
     matz = _STATE_VAR_S_(data%id_sedzone)  ! local benthic type
     mpb  = _STATE_VAR_S_(data%id_mpb)      ! local mpb density
     Io   = _STATE_VAR_S_(data%id_I_0)      ! surface short wave radiation
     temp = _STATE_VAR_(data%id_tem)        ! local temperature
     extc = _STATE_VAR_(data%id_extc)       ! cell extinction
     par  = _STATE_VAR_(data%id_par)        ! local photosynt. active radiation
     dz   = _STATE_VAR_(data%id_dz)         ! cell depth

     ! Get sedimentation flux (mmmol/m2/d) loss into the benthos (diagnostic was set in mobility)
     Psed_phy = _DIAG_VAR_(data%id_Psed_phy)

     ! Compute photosynthesis and respiration
     fI = photosynthesis_irradiance(3,data%I_Kmpb,data%I_Kmpb,par,extc,Io,dz)
     mpb_prod = data%R_mpbg*fI*(data%theta_mpb_growth**(temp-20.))*(1.-(MIN(mpb,data%mpb_max)/data%mpb_max))
     mpb_resp = (data%R_mpbr*(data%theta_mpb_resp**(temp-20.)))           !*(( (mpb-mpb_min)/(data%mpb_max-mpb_min) )
     mpb_flux = (mpb_prod-mpb_resp)*mpb

     ! Update the MPB biomass, include net production and add sedimented phytos (mmolC/m2/s)
     _FLUX_VAR_B_(data%id_mpb) = _FLUX_VAR_B_(data%id_mpb) + mpb_flux + (-Psed_phy/secs_per_day)

     ! log this uptake into the bulk community GPP/NCP diagnostics (mmolC/m3/d)
     _DIAG_VAR_(data%id_GPP) =  _DIAG_VAR_(data%id_GPP) + (mpb_prod/dz) * mpb * secs_per_day
     _DIAG_VAR_(data%id_NCP) =  _DIAG_VAR_(data%id_NCP) + (mpb_flux/dz) * secs_per_day

     ! Update flux terms for other O2, CO2 and nutrient fluxes (mmolO2/m2/s)
     IF (data%do_DOuptake) THEN
        _FLUX_VAR_(data%id_DOupttarget) = _FLUX_VAR_(data%id_DOupttarget) + mpb_flux
     ENDIF
     IF (data%do_Cuptake) THEN
        _FLUX_VAR_(data%id_Cupttarget) = _FLUX_VAR_(data%id_Cupttarget) - mpb_flux
        ! log this uptake into the bulk community C uptake diagnostic
        _DIAG_VAR_(data%id_CUP) = _DIAG_VAR_(data%id_CUP) - (mpb_flux/dz) * secs_per_day
     ENDIF
     ! A quick and dirty nutrient uptake by MPB; needs cleaning to account for limitation and excretion of DOM
     IF (data%do_Nuptake) THEN
        ! increment flux from bottom cell nutrient amount (mmol N/m2/s)
        _FLUX_VAR_(data%id_Nupttarget(1)) = &
                           _FLUX_VAR_(data%id_Nupttarget(1)) - mpb_flux * (16./106.) *0.5
        _FLUX_VAR_(data%id_Nupttarget(2)) = &
                           _FLUX_VAR_(data%id_Nupttarget(2)) - mpb_flux * (16./106.) *0.5
        ! log this uptake into the bulk community N uptake diagnostic (mmol N/m3/d)
        _DIAG_VAR_(data%id_NUP1)= _DIAG_VAR_(data%id_NUP1)- (mpb_flux/dz) * (16./106.) *0.5 * secs_per_day
        _DIAG_VAR_(data%id_NUP2)= _DIAG_VAR_(data%id_NUP2)- (mpb_flux/dz) * (16./106.) *0.5 * secs_per_day
     ENDIF
     IF (data%do_Puptake) THEN
        ! increment flux from bottom cell nutrient (po4) amount (mmol P/m2/s)
        _FLUX_VAR_(data%id_Pupttarget(1)) = _FLUX_VAR_(data%id_Pupttarget(1)) - mpb_flux * (1./106.)
        ! log this uptake into the bulk community P uptake diagnostic (mmol P/m3/d)
        _DIAG_VAR_(data%id_PUP) = _DIAG_VAR_(data%id_PUP) - mpb_flux * (1./106.) * secs_per_day
     ENDIF

     ! Now compute resuspension of mpb and distribute into the phytoplankton groups
     Fsed_phy = zero_
     IF ( data%n_zones > 0 ) THEN
       IF( in_zone_set(matz,data%active_zones) .AND. data%id_l_resus > 0 ) THEN

         phy_sed_frac = SUM(data%resuspension(1:data%num_phytos))
         Fsed_phy = _DIAG_VAR_S_(data%id_l_resus) * phy_sed_frac

         _FLUX_VAR_B_(data%id_mpb) = _FLUX_VAR_B_(data%id_mpb) - Fsed_phy

         IF( phy_sed_frac > zero_ ) THEN
          DO phy_i=1,data%num_phytos
           ! spread the resuspended phytoplantkon mass between the groups based
           ! on the relative sed fractions provided by user in "resuspension(:)"
           ! note, this is incremented into the SED diag for the group in this
           ! bottom layer
           resus = Fsed_phy * (data%resuspension(phy_i) / phy_sed_frac)

           _FLUX_VAR_(data%id_p(phy_i)) = _FLUX_VAR_(data%id_p(phy_i)) + resus
           _DIAG_VAR_(data%id_PhySEDc(phy_i)) = _DIAG_VAR_(data%id_PhySEDc(phy_i)) + resus/dz * secs_per_day
                                                        ! check if incrementing here is right order (after mobility)

           IF(data%phytos(phy_i)%simINDynamics>0) THEN
             _FLUX_VAR_(data%id_in(phy_i)) = _FLUX_VAR_(data%id_in(phy_i)) + resus * (16./106.)
             _DIAG_VAR_(data%id_PhySEDn(phy_i)) = _DIAG_VAR_(data%id_PhySEDn(phy_i)) + &
                                                                       resus/dz * (16./106.) * secs_per_day
           ENDIF
           IF(data%phytos(phy_i)%simIPDynamics>0) THEN
             _FLUX_VAR_(data%id_ip(phy_i)) = _FLUX_VAR_(data%id_ip(phy_i)) + resus * (1./106.)
            _DIAG_VAR_(data%id_PhySEDp(phy_i)) = _DIAG_VAR_(data%id_PhySEDp(phy_i)) + &
                                                                       resus/dz * (1./106.) * secs_per_day
           ENDIF
          ENDDO
         ENDIF
       ENDIF
     ENDIF

     ! Update the diagnostic variables (mmol C/m2/d)
     _DIAG_VAR_S_(data%id_d_mpb) = mpb
     _DIAG_VAR_S_(data%id_d_bpp) =      (mpb_prod) * mpb * secs_per_day
     _DIAG_VAR_S_(data%id_d_bcp) =              mpb_flux * secs_per_day
     _DIAG_VAR_S_(data%id_d_mpbv)= -Psed_phy - (Fsed_phy * secs_per_day)
     _DIAG_VAR_S_(data%id_d_res) =              Fsed_phy * secs_per_day
   ENDIF
END SUBROUTINE aed_calculate_benthic_phytoplankton
!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


!###############################################################################
SUBROUTINE aed_mobility_phytoplankton(data,column,layer_idx,mobility)
!------------------------------------------------------------------------------+
! Get the vertical movement values for phytoplankton cells
!------------------------------------------------------------------------------+
!ARGUMENTS
   CLASS (aed_phytoplankton_data_t),INTENT(in) :: data
   TYPE (aed_column_t),INTENT(inout) :: column(:)
   INTEGER,INTENT(in) :: layer_idx
   AED_REAL,INTENT(inout) :: mobility(:)
!
!LOCALS
   AED_REAL :: temp, par, rho_p, Io, dz
   AED_REAL :: vvel
   AED_REAL :: pw, pw20, mu, mu20
   AED_REAL :: IN, IC, Q, Qmax
   INTEGER  :: phy_i
!
!------------------------------------------------------------------------------+
!BEGIN
   _DIAG_VAR_(data%id_Psed_phy) = zero_

   dz = _STATE_VAR_(data%id_dz)

   DO phy_i=1,data%num_phytos
      SELECT CASE (data%phytos(phy_i)%settling)

         CASE ( _MOB_OFF_ )
            ! disable settling by setting vertical velocity to 0
            vvel = zero_

         CASE ( _MOB_CONST_ )
            ! constant settling velocity using user provided value
            vvel = data%phytos(phy_i)%w_p

         CASE ( _MOB_TEMP_ )
            ! constant settling velocity @20C corrected for density changes
            pw   = _STATE_VAR_(data%id_dens)
            temp = _STATE_VAR_(data%id_tem)
            mu   = water_viscosity(temp)
            mu20 = 0.001002  ! N s/m2
            pw20 = 998.2000  ! kg/m3 (assuming freshwater)
            vvel = data%phytos(phy_i)%w_p*mu20*pw / ( mu*pw20 )

         CASE ( _MOB_STOKES_ )
            ! settling velocity based on Stokes Law calculation and cell density
            pw   = _STATE_VAR_(data%id_dens)           ! water density
            temp = _STATE_VAR_(data%id_tem)
            mu   = water_viscosity(temp)               ! water dynamic viscosity
            IF( data%id_rho(phy_i)>0 ) THEN
              rho_p = _STATE_VAR_(data%id_rho(phy_i))  ! cell density
            ELSE
              rho_p = data%phytos(phy_i)%rho_phy
            ENDIF
            vvel = -9.807*(data%phytos(phy_i)%d_phy**2.)*( rho_p-pw ) / ( 18.*mu )

          CASE ( _MOB_MOTILE_ )
             ! vertical velocity based on motility and behaviour of phyto group
             ! modelled as in Ross and Sharples (2007)
             par = _STATE_VAR_(data%id_par)     ! local photosynth. active radiation
             Io  = _STATE_VAR_S_(data%id_I_0)   ! surface short wave radiation

             vvel = zero_
             IC = _STATE_VAR_(data%id_p(phy_i))
             IN = _STATE_VAR_(data%id_in(phy_i))
             Q = IN/IC
             Qmax = data%phytos(phy_i)%X_nmax
             vvel = zero_
             IF( Q<(data%phytos(phy_i)%f1*Qmax) ) THEN
               !-- nutrient starvation, swim down
               vvel = -data%phytos(phy_i)%w_p
             ELSEIF(Q>(data%phytos(phy_i)%f2*Qmax)) THEN
               !-- nutrient replete
               IF( par > data%phytos(phy_i)%I_K ) THEN
                 !-- high light, swim up
                 vvel = data%phytos(phy_i)%w_p
               ELSE
                 !-- low light
                 vvel = zero_
               ENDIF
            ELSE
               vvel = zero_
            ENDIF
            IF (par>0.95*Io) vvel = zero_   ! already at the surface

         CASE DEFAULT
            ! unknown settling/migration option selection
            vvel =  zero_

      END SELECT

      ! set global mobility array for phytoplankton group
      mobility(data%id_p(phy_i)) = vvel
      IF( data%id_rho(phy_i)>0 ) mobility(data%id_rho(phy_i)) = vvel
      IF( diag_level >= 10  .AND. data%id_vvel(phy_i)>0) &
              _DIAG_VAR_(data%id_vvel(phy_i)) = vvel * secs_per_day  ! m/day

      ! record carbon sedimentation flux (mmol C /m3 /day)
      _DIAG_VAR_(data%id_PhySEDc(phy_i)) = (vvel/dz) * _STATE_VAR_(data%id_p(phy_i)) * secs_per_day

      ! set global mobility array for phytoplankton IN pool, and record flux
      IF(data%phytos(phy_i)%simINDynamics>0) THEN
         mobility(data%id_in(phy_i)) = vvel
         _DIAG_VAR_(data%id_PhySEDn(phy_i)) = (vvel/dz) * _STATE_VAR_(data%id_in(phy_i))* secs_per_day
      ELSE
         _DIAG_VAR_(data%id_PhySEDn(phy_i)) = (vvel/dz) * _STATE_VAR_(data%id_p(phy_i))* &
                                                               data%phytos(phy_i)%X_ncon* secs_per_day
      ENDIF
      ! set global mobility array for phytoplankton IP pool, and record flux
      IF(data%phytos(phy_i)%simIPDynamics>0) THEN
         mobility(data%id_ip(phy_i)) = vvel
         _DIAG_VAR_(data%id_PhySEDp(phy_i)) = (vvel/dz) * _STATE_VAR_(data%id_ip(phy_i))* secs_per_day
      ELSE
         _DIAG_VAR_(data%id_PhySEDp(phy_i)) = (vvel/dz) * _STATE_VAR_(data%id_p(phy_i))* &
                                                               data%phytos(phy_i)%X_pcon* secs_per_day
      ENDIF

       ! cumulate the community sedimentation flux (mmmol/m2/d) for later use
       _DIAG_VAR_(data%id_Psed_phy) =  _DIAG_VAR_(data%id_Psed_phy)  &
                           + vvel * _STATE_VAR_(data%id_p(phy_i)) * secs_per_day

    ENDDO
END SUBROUTINE aed_mobility_phytoplankton
!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


!###############################################################################
SUBROUTINE aed_light_extinction_phytoplankton(data,column,layer_idx,extinction)
!------------------------------------------------------------------------------+
! Get the light extinction coefficient due to biogeochemical variables
!------------------------------------------------------------------------------+
!ARGUMENTS
   CLASS (aed_phytoplankton_data_t),INTENT(in) :: data
   TYPE (aed_column_t),INTENT(inout) :: column(:)
   INTEGER,INTENT(in) :: layer_idx
   AED_REAL,INTENT(inout) :: extinction
!
!LOCALS
   AED_REAL :: phy
   INTEGER  :: phy_i
!
!------------------------------------------------------------------------------+
!BEGIN

   DO phy_i=1,data%num_phytos
      ! Retrieve current (local) state variable values.
      phy = _STATE_VAR_(data%id_p(phy_i))! phytoplankton

      ! Self-shading with contribution from this phytoplankton concentration.
      extinction = extinction + (data%phytos(phy_i)%KePHY*phy)
   ENDDO
END SUBROUTINE aed_light_extinction_phytoplankton
!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


END MODULE aed_phytoplankton
